"""
This module contains all functions which are used to improve the documentation of attributes.
"""
from django.db import models
from django.db.models.fields import related_descriptors
from django.db.models.fields.files import FileDescriptor
from django.db.models.manager import ManagerDescriptor
from django.db.models.query_utils import DeferredAttribute
from django.utils.module_loading import import_string
from sphinx.util.docstrings import prepare_docstring

from .config import CHOICES_LIMIT
from .field_utils import get_field_type, get_field_verbose_name

FIELD_DESCRIPTORS = (FileDescriptor, related_descriptors.ForwardManyToOneDescriptor)

# Support for some common third party fields
try:
    from phonenumber_field.modelfields import PhoneNumberDescriptor

    FIELD_DESCRIPTORS += (PhoneNumberDescriptor,)
except ImportError:
    PhoneNumberDescriptor = None


def improve_attribute_docstring(attribute, name, lines):
    """
    Improve the documentation of various model fields.

    This improves the navigation between related objects.

    :param attribute: The instance of the object to document
    :type attribute: object

    :param name: The full dotted path to the object
    :type name: str

    :param lines: The docstring lines
    :type lines: list [ str ]
    """
    # Save initial docstring lines to append them to the modified lines
    docstring_lines = lines.copy()
    lines.clear()
    if isinstance(attribute, DeferredAttribute):
        # This only points to a field name, not a field.
        # Get the field by importing the name.
        cls_path, field_name = name.rsplit(".", 1)
        model = import_string(cls_path)
        field = model._meta.get_field(field_name)
        if isinstance(field, models.fields.related.RelatedField):
            # If a deferred attribute is a related field, it is an automatically created field
            # with the postfix "_id" and contains the reference to the id of the related model
            # instance. These are usually undocumented, so they only are included in the docs
            # is sphinx is invoked with the undoc-members option.
            lines.append(
                f"Internal field, use :class:`~{cls_path}.{field.name}` instead."
            )
        else:
            lines.extend(get_field_details(field))
    elif isinstance(attribute, FIELD_DESCRIPTORS):
        # Display a reasonable output for forward descriptors (foreign key and one to one fields).
        lines.extend(get_field_details(attribute.field))
    elif isinstance(attribute, related_descriptors.ManyToManyDescriptor):
        # Check this case first since ManyToManyDescriptor inherits from ReverseManyToOneDescriptor
        # This descriptor is used for both forward and reverse relationships
        if attribute.reverse:
            lines.extend(get_field_details(attribute.rel))
        else:
            lines.extend(get_field_details(attribute.field))
    elif isinstance(attribute, related_descriptors.ReverseManyToOneDescriptor):
        lines.extend(get_field_details(attribute.rel))
    elif isinstance(attribute, related_descriptors.ReverseOneToOneDescriptor):
        lines.extend(get_field_details(attribute.related))
    elif isinstance(attribute, (models.Manager, ManagerDescriptor)):
        # Somehow the 'objects' manager doesn't pass through the docstrings.
        module, model_name, field_name = name.rsplit(".", 2)
        lines.append("Django manager to access the ORM")
        lines.append(f"Use ``{model_name}.objects.all()`` to fetch all objects.")
    # Check if there are initial docstrings to be appended
    if docstring_lines:
        # Get default docstring of attribute
        parent_docstring = type(attribute).__doc__
        # Ignore non-string __doc__
        if not isinstance(parent_docstring, str):
            parent_docstring = ""
        # Only append the initial docstring of the attribute if it's overwritten
        if docstring_lines != prepare_docstring(parent_docstring) or not lines:
            if lines:
                # If lines are not empty, append a separating new line before docstring
                lines.append("")
            # Remove last element because it's a newline
            lines.extend(docstring_lines[:-1])


def get_field_details(field):
    """
    This function returns the detail docstring of a model field.
    It includes the field type and the verbose name of the field.

    :param field: The field
    :type field: ~django.db.models.Field

    :return: The field details as list of strings
    :rtype: list [ str ]
    """
    field_details = [
        f"Type: {get_field_type(field)}",
        "",
        f"{get_field_verbose_name(field)}",
    ]
    if hasattr(field, "choices") and field.choices:
        field_details.extend(["", "Choices:", ""])
        field_details.extend(
            [f"* ``{key}``" for key, value in field.choices[:CHOICES_LIMIT]]
        )
        # Check if list has been truncated
        if len(field.choices) > CHOICES_LIMIT:
            # If only one element has been truncated, just list it as well
            if len(field.choices) == CHOICES_LIMIT + 1:
                field_details.append(f"* ``{field.choices[-1][0]}``")
            else:
                field_details.append(f"* and {len(field.choices) - CHOICES_LIMIT} more")
    return field_details



"""
This module contains patches for Django to improve its interaction with Sphinx.
"""
from django import apps, forms, test
from django.db import models

try:
    from django.contrib.postgres import fields as postgres_fields
    from django.contrib.postgres import forms as postgres_forms

    POSTGRES = True
except ModuleNotFoundError:
    # In case postgres is not used, pass
    POSTGRES = False

try:
    from mptt.managers import TreeManager

    MPTT = True
except ModuleNotFoundError:
    # In case postgres is not used, pass
    MPTT = False


def patch_django_for_autodoc():
    """
    Fix the appearance of some classes in autodoc.
    E.g. the absolute path to the base model class is ``django.db.models.base.Model``, but its
    intersphinx mapping path is ``django.db.models.Model``.

    This also avoids query evaluation.
    """
    # Fix Django's manager appearance
    models.manager.ManagerDescriptor.__get__ = (
        lambda self, *args, **kwargs: self.manager
    )

    # Stop Django from executing DB queries
    models.QuerySet.__repr__ = lambda self: self.__class__.__name__

    # Fix django-mptt TreeManager in Django >=3.1
    if MPTT:
        TreeManager.get_queryset = models.Manager.get_queryset

    # Module paths which are documented in the parent module
    DJANGO_MODULE_PATHS = {
        "django.db.models": [
            models.base,
            models.fields,
            models.fields.files,
            models.fields.related,
        ],
        "django.forms": [forms.forms, forms.fields, forms.models, forms.widgets],
        "django.test": [test],
        "django.apps": [apps],
    }

    # Support django.db.models.JSONField in Django >= 3.1
    if hasattr(models.fields, "json"):
        DJANGO_MODULE_PATHS["django.db.models"].append(models.fields.json)

    # Support postgres fields if used
    if POSTGRES:
        DJANGO_MODULE_PATHS["django.contrib.postgres.forms"] = [
            postgres_forms.array,
            postgres_forms.jsonb,
        ]
        DJANGO_MODULE_PATHS["django.contrib.postgres.fields"] = [
            postgres_fields.array,
            postgres_fields.jsonb,
        ]
        postgres_forms.array.__all__ = ("SimpleArrayField", "SplitArrayField")
        postgres_forms.jsonb.__all__ = ("JSONString", "JSONField")

    # Add __all__ where missing
    models.base.__all__ = ("Model", "FilteredRelation")
    models.fields.files.__all__ = ("FileField", "ImageField")
    models.fields.related.__all__ = ("ForeignKey", "OneToOneField", "ManyToManyField")

    # Set the __module__ to the parent module to make sure intersphinx mappings work as expected
    for parent_module_str, django_modules in DJANGO_MODULE_PATHS.items():
        for django_module in django_modules:
            for module_class in map(django_module.__dict__.get, django_module.__all__):
                try:
                    module_class.__module__ = parent_module_str
                except AttributeError:
                    pass

    # Fix module path of model manager
    models.manager.Manager.__module__ = "django.db.models"



"""
This is a sphinx extension which improves the documentation of Django apps.
"""
__version__ = "1.3"

from . import docstrings, roles


def setup(app):
    """
    Allow this module to be used as sphinx extension.

    Setup the two sub-extensions :mod:`~sphinxcontrib_django2.docstrings` and
    :mod:`~sphinxcontrib_django2.roles` which can also be imported separately.

    :param app: The Sphinx application object
    :type app: ~sphinx.application.Sphinx
    """
    docstrings.setup(app)
    roles.setup(app)

    return {
        "version:": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""
This module contains all functions which are used to improve the documentation of methods.
"""
import re

RE_GET_FOO_DISPLAY = re.compile(r"\.get_(?P<field>[a-zA-Z0-9_]+)_display$")
RE_GET_NEXT_BY = re.compile(r"\.get_next_by_(?P<field>[a-zA-Z0-9_]+)$")
RE_GET_PREVIOUS_BY = re.compile(r"\.get_previous_by_(?P<field>[a-zA-Z0-9_]+)$")


def improve_method_docstring(name, lines):
    """
    Improve the documentation of methods automatically contributed to models by Django:

    * :meth:`~django.db.models.Model.get_FOO_display`
    * :meth:`~django.db.models.Model.get_next_by_FOO`
    * :meth:`~django.db.models.Model.get_previous_by_FOO`

    :param name: The full dotted path to the object.
    :type name: str

    :param lines: The lines of docstring lines
    :type lines: list [ str ]
    """
    if not lines:
        # Not doing obj.__module__ lookups to avoid performance issues.
        if name.endswith("_display"):
            match = RE_GET_FOO_DISPLAY.search(name)
            if match is not None:
                # Django get_..._display method
                lines.append(
                    f"Shows the label of the :attr:`{match.group('field')}`. "
                    f"See :meth:`~django.db.models.Model.get_FOO_display` for more information."
                )
        elif ".get_next_by_" in name:
            match = RE_GET_NEXT_BY.search(name)
            if match is not None:
                lines.append(
                    f"Finds next instance based on :attr:`{match.group('field')}`. "
                    f"See :meth:`~django.db.models.Model.get_next_by_FOO` for more information."
                )
        elif ".get_previous_by_" in name:
            match = RE_GET_PREVIOUS_BY.search(name)
            if match is not None:
                lines.append(
                    f"Finds previous instance based on :attr:`{match.group('field')}`. "
                    f"See :meth:`~django.db.models.Model.get_previous_by_FOO` for more "
                    f"information."
                )



"""
This module contains all functions which are used to improve the documentation of classes.
"""

from django import forms
from django.db import models

from .field_utils import get_field_type, get_field_verbose_name


def improve_class_docstring(app, cls, lines):
    """
    Improve the documentation of a class if it's a Django model or form

    :param app: The Sphinx application object
    :type app: ~sphinx.application.Sphinx

    :param cls: The instance of the class to document
    :type cls: object

    :param lines: The docstring lines
    :type lines: list [ str ]
    """
    if issubclass(cls, models.Model):
        improve_model_docstring(app, cls, lines)
    elif issubclass(cls, forms.BaseForm):
        improve_form_docstring(cls, lines)


def improve_model_docstring(app, model, lines):
    """
    Improve the documentation of a Django :class:`~django.db.models.Model` subclass.

    This adds all model fields as parameters to the ``__init__()`` method.

    :param app: The Sphinx application object
    :type app: ~sphinx.application.Sphinx

    :param model: The instance of the model to document
    :type model: ~django.db.models.Model

    :param lines: The docstring lines
    :type lines: list [ str ]
    """

    # Add database table name
    if app.config.django_show_db_tables:
        lines.insert(0, "")
        lines.insert(0, f"**Database table:** ``{model._meta.db_table}``")

    # Get predefined params to exclude them from the automatically inserted params
    param_offset = len(":param ")
    predefined_params = [
        line[param_offset : line.find(":", param_offset)]
        for line in lines
        if line.startswith(":param ") and ":" in line[param_offset:]
    ]

    # Get all fields of this model which are not already explicitly included in the docstring
    all_fields = [
        field
        for field in model._meta.get_fields(include_parents=True)
        if field.name not in predefined_params
    ]
    # Get all related fields (ForeignKey, OneToOneField, ManyToManyField)
    related_fields = [
        field
        for field in all_fields
        if isinstance(field, models.fields.related.RelatedField)
    ]
    # Get all reverse relationships
    reverse_related_fields = [
        field
        for field in all_fields
        if isinstance(field, models.fields.reverse_related.ForeignObjectRel)
    ]
    # All fields which are neither related nor reverse related
    non_related_fields = [
        field
        for field in all_fields
        if field not in related_fields + reverse_related_fields
    ]

    # Add the normal fields to the docstring
    add_model_parameters(non_related_fields, lines)

    # Add the related fields
    if related_fields:
        lines.append("")
        lines.append("Relationship fields:")
        lines.append("")
        add_model_parameters(related_fields, lines)

    # Add the reverse related fields
    if reverse_related_fields:
        lines.append("")
        lines.append("Reverse relationships:")
        lines.append("")
        add_model_parameters(reverse_related_fields, lines)

    # Add the inheritance diagram
    if (
        "sphinx.ext.inheritance_diagram" in app.extensions
        and "sphinx.ext.graphviz" in app.extensions
        and not any("inheritance-diagram::" in line for line in lines)
    ):
        lines.append(".. inheritance-diagram::")  # pragma: no cover


def add_model_parameters(fields, lines):
    """
    Add the given fields as model parameter with the ``:param:`` directive

    :param fields: The list of fields
    :type fields: list [ ~django.db.models.Field ]

    :param lines: The list of current docstring lines
    :type lines: list [ str ]
    """
    for field in fields:
        # Add verbose name
        lines.append(f":param {field.name}: {get_field_verbose_name(field)}")

        # Add type
        lines.append(f":type {field.name}: {get_field_type(field, include_role=False)}")


def improve_form_docstring(form, lines):
    """
    Improve the documentation of a Django :class:`~django.forms.Form` class.
    This highlights the available fields in the form.

    :param form: The form object
    :type form: ~django.forms.Form

    :param lines: The list of existing docstring lines
    :type lines: list [ str ]
    """
    lines.append("**Form fields:**")
    lines.append("")
    for name, field in form.base_fields.items():
        field_type = f"{field.__class__.__module__}.{field.__class__.__name__}"
        label = field.label or name.replace("_", " ").title()
        lines.append(f"* ``{name}``: {label} (:class:`~{field_type}`)")



import io
import sys

from pprintpp import pprint as pp


def improve_data_docstring(data, lines):
    """
    Improve the documentation of data by pretty-printing into in the docstring.

    :param data: The documented object
    :type data: object

    :param lines: The lines of docstring lines
    :type lines: list [ str ]
    """
    if isinstance(data, (list, tuple, dict, set)):
        # Redirect stdout to StringIO to catch print
        old_stdout = sys.stdout
        new_stdout = io.StringIO()
        sys.stdout = new_stdout

        # Pretty print iterable
        pp(data)
        output = new_stdout.getvalue()

        # Append pretty printed lines
        lines.append(".. code-block:: JavaScript")
        lines.append("")
        lines.append("    " + output)

        # Reset stdout
        sys.stdout = old_stdout



"""
This module contains utiliy functions for fields which are used by both the
:mod:`~sphinxcontrib_django2.docstrings.attributes` and
:mod:`~sphinxcontrib_django2.docstrings.classes` modules.
"""
from django.apps import apps
from django.contrib import contenttypes
from django.db import models
from django.utils.encoding import force_str


def get_field_type(field, include_role=True):
    """
    Get the type of a field including the correct intersphinx mappings.

    :param field: The field
    :type field: ~django.db.models.Field

    :param include_directive: Whether or not the role :any:`py:class` should be included
    :type include_directive: bool

    :return: The type of the field
    :rtype: str
    """
    if isinstance(field, models.fields.related.RelatedField):
        if isinstance(field.remote_field.model, str):
            # This happens with foreign keys of abstract models
            to = field.remote_field.model
        else:
            to = f"{field.remote_field.model.__module__}.{field.remote_field.model.__name__}"
        return f":class:`~{type(field).__module__}.{type(field).__name__}` to :class:`~{to}`"
    elif isinstance(field, models.fields.reverse_related.ForeignObjectRel):
        to = field.remote_field.model
        return (
            f"Reverse :class:`~{type(field.remote_field).__module__}."
            f"{type(field.remote_field).__name__}` from :class:`~{to.__module__}.{to.__name__}`"
        )
    else:
        if include_role:
            # For the docstrings of attributes, the :class: role is required
            return f":class:`~{type(field).__module__}.{type(field).__name__}`"
        else:
            # For the :param: role in class docstrings, the :class: role is not required
            return f"~{type(field).__module__}.{type(field).__name__}"


def get_field_verbose_name(field):
    """
    Get the verbose name of the field.
    If the field has a ``help_text``, it is also included.

    In case the field is a related field, the ``related_name`` is used to link to the remote model.
    For reverse related fields, the originating field is linked.

    :param field: The field
    :type field: ~django.db.models.Field
    """
    help_text = ""
    # Check whether the field is a reverse related field
    if isinstance(field, models.fields.reverse_related.ForeignObjectRel):
        # Convert related name to a readable name if ``snake_case`` is used
        related_name = (
            field.related_name.replace("_", " ") if field.related_name else None
        )
        if isinstance(field, models.fields.reverse_related.OneToOneRel):
            # If a related name is given, use it, else use the verbose name of the remote model
            related_name = related_name or field.remote_field.model._meta.verbose_name
            # If field is a OneToOne field, use the prefix "The"
            verbose_name = (
                f"The {related_name} of this {field.model._meta.verbose_name}"
            )
        else:
            # This means field is an instance of ManyToOneRel or ManyToManyRel
            # If a related name is given, use it, else use the verbose name of the remote model
            related_name = (
                related_name or field.remote_field.model._meta.verbose_name_plural
            )
            # If field is a foreign key or a ManyToMany field, use the prefix "All"
            verbose_name = (
                f"All {related_name} of this {field.model._meta.verbose_name}"
            )
        # Always link to the origin of the reverse related field
        verbose_name += (
            f" (related name of :attr:`~{field.remote_field.model.__module__}"
            f".{field.remote_field.model.__name__}.{field.remote_field.name}`)"
        )
    elif hasattr(contenttypes, "fields") and isinstance(
        field, contenttypes.fields.GenericForeignKey
    ):
        # GenericForeignKey does not inherit from django.db.models.Field and has no verbose_name
        return (
            f"Generic foreign key to the :class:`~django.contrib.contenttypes.models.ContentType` "
            f"specified in "
            f":attr:`~{field.model.__module__}.{field.model.__name__}.{field.ct_field}`"
        )
    else:
        # This means the field is either a normal field or a forward related field
        # If the field is a primary key, include a notice
        primary_key = "Primary key: " if field.primary_key else ""

        field_verbose_name = force_str(field.verbose_name)
        # Make the first letter upper case while leave the rest unchanged
        # (str.capitalize() would make the rest lower case, e.g. ID => Id)
        verbose_name = (
            primary_key + field_verbose_name[:1].upper() + field_verbose_name[1:]
        )
        help_text = force_str(field.help_text)

    # Add help text if field has one
    if help_text:
        # Separate verbose name and help text by a dot
        if not verbose_name.endswith("."):
            verbose_name += ". "
        verbose_name += help_text

    if isinstance(field, models.fields.related.RelatedField):
        # If field is a forward related field, reference the remote model
        to = field.remote_field.model
        if isinstance(to, str):
            # This happens with foreign keys of abstract models
            if "." in to:
                to = apps.get_model(to)
            elif to == "self":
                to = field.model
            else:
                to = apps.get_model(field.model._meta.app_label, to)
        # If a related name is defined
        if hasattr(field.remote_field, "related_name"):
            related_name = (
                field.remote_field.related_name or field.model.__name__.lower()
            )

        verbose_name += (
            f" (related name: :attr:`~{to.__module__}.{to.__name__}.{related_name}`)"
        )
    return verbose_name



"""
This module contains configuration of the members which should in-/excluded in sphinx
(see :event:`autodoc-skip-member`)
"""
#: Ensure that the __init__ method gets documented (also see :confval:`autoclass_content` setting)
INCLUDE_MEMBERS = {"__init__"}

#: Members to hide.
EXCLUDE_MEMBERS = {
    # BaseForm
    "base_fields",
    "declared_fields",
    "Meta",
    # BaseModelAdmin
    "declared_fieldsets",
    "fieldsets",
    # Wagtail Page
    "panels",
    "content_panels",
    # Polymorphic
    "polymorphic_primary_key_name",
    "polymorphic_super_sub_accessors_replaced",
}

#: How many choices should be shown for model fields
CHOICES_LIMIT = 10



#!/usr/bin/env python
import codecs
import re
from os import path

from setuptools import find_packages, setup


def read(*parts):
    file_path = path.join(path.dirname(__file__), *parts)
    return codecs.open(file_path, encoding="utf-8").read()


def find_version(*parts):
    version_file = read(*parts)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return str(version_match.group(1))
    raise RuntimeError("Unable to find version string.")


setup(
    name="sphinxcontrib-django2",
    version=find_version("sphinxcontrib_django2", "__init__.py"),
    license="Apache 2.0",
    description="Improve the Sphinx autodoc for Django classes.",
    long_description=read("README.rst"),
    author="Timo Ludwig",
    author_email="ti.ludwig@web.de",
    url="https://github.com/timoludwig/sphinxcontrib-django2",
    download_url="https://github.com/timoludwig/sphinxcontrib-django2/zipball/main",
    packages=find_packages(exclude=("example*",)),
    install_requires=["Django>=2.2", "Sphinx>=0.5", "pprintpp"],
    extras_require={
        "dev": ["pre-commit"],
        "test": ["pytest", "requests-mock", "codecov"],
        "doc": ["sphinx-rtd-theme", "sphinx-last-updated-by-git"],
        "optional": [
            "psycopg2-binary",
            "django-mptt",
            "django-phonenumber-field[phonenumbers]",
        ],
    },
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



"""
This module adds cross-reference types which are used in the documentation of Django and Sphinx to
allow intersphinx mappings to these custom types.
The supported text roles are:

Django:

* ``:setting:``, e.g. ``:setting:`INSTALLED_APPS``` renders as :setting:`INSTALLED_APPS`
* ``:templatetag:``, e.g. ``:templatetag:`block``` renders as :templatetag:`block`
* ``:templatefilter:``, e.g. ``:templatefilter:`add``` renders as :templatefilter:`add`
* ``:fieldlookup:``, e.g. ``:fieldlookup:`equals``` renders as :fieldlookup:`equals`

Sphinx:

* ``:event:``, e.g. ``:event:`autodoc-skip-member``` renders as :event:`autodoc-skip-member`
* ``:confval:``, e.g. ``:confval:`extensions``` renders as :confval:`extensions`

This module can also be used separately in ``conf.py``::

    extensions = [
        "sphinxcontrib_django2.roles",
    ]
"""
import logging

from sphinx.errors import ExtensionError

from . import __version__

logger = logging.getLogger(__name__)


def setup(app):
    """
    Allow this module to be used as Sphinx extension.

    This is also called from the top-level :meth:`~sphinxcontrib_django2.setup`.

    It adds cross-reference types via :meth:`~sphinx.application.Sphinx.add_crossref_type`.

    :param app: The Sphinx application object
    :type app: ~sphinx.application.Sphinx
    """
    # Load sphinx.ext.intersphinx extension
    app.setup_extension("sphinx.ext.intersphinx")

    # Add default intersphinx mappings after config is initialized
    app.connect("config-inited", add_default_intersphinx_mappings)

    # Allow intersphinx mappings to custom Django roles
    django_crossref_types = ["setting", "templatetag", "templatefilter", "fieldlookup"]
    # Allow intersphinx mappings to custom Sphinx roles
    sphinx_crossref_types = ["event", "confval"]

    for crossref_type in django_crossref_types + sphinx_crossref_types:
        try:
            app.add_crossref_type(directivename=crossref_type, rolename=crossref_type)
        except ExtensionError as e:
            logger.warning("Unable to register cross-reference type: %s", e)

    return {
        "version:": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


def add_default_intersphinx_mappings(app, config):
    """
    This function provides a default intersphinx mapping to the documentations of Python, Django
    and Sphinx if ``intersphinx_mapping`` is not given in ``conf.py``.

    Called on the :event:`config-inited` event.

    :param app: The Sphinx application object
    :type app: ~sphinx.application.Sphinx

    :param config: The Sphinx configuration
    :type config: ~sphinx.config.Config
    """
    DEFAULT_INTERSPHINX_MAPPING = {
        "python": ("https://docs.python.org/", None),
        "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
        "django": (
            "https://docs.djangoproject.com/en/stable/",
            "https://docs.djangoproject.com/en/stable/_objects/",
        ),
    }
    if not config.intersphinx_mapping:
        config.intersphinx_mapping = DEFAULT_INTERSPHINX_MAPPING



"""
Improve the docstrings of Django apps.

For example:

* List all model and form fields as parameters
  (see:mod:`~sphinxcontrib_django2.docstrings.classes`)
* Improve field representations in the documentation
  (see :mod:`~sphinxcontrib_django2.docstrings.attributes`)
* Add information about autogenerated methods
  (see :mod:`~sphinxcontrib_django2.docstrings.methods`)
* Improve the appearance of static iterable data
  (see :mod:`~sphinxcontrib_django2.docstrings.data`)
* Fix the intersphinx mappings to the Django documentation
  (see :mod:`~sphinxcontrib_django2.docstrings.patches`)
"""
import importlib
import os

import django
from sphinx.errors import ConfigError

from .. import __version__
from .attributes import improve_attribute_docstring
from .classes import improve_class_docstring
from .config import EXCLUDE_MEMBERS, INCLUDE_MEMBERS
from .data import improve_data_docstring
from .methods import improve_method_docstring


def setup(app):
    """
    Allow this package to be used as Sphinx extension.

    This is also called from the top-level :meth:`~sphinxcontrib_django2.setup`.

    It connects to the sphinx events :event:`autodoc-skip-member` and
    :event:`autodoc-process-docstring`.

    Additionally, the sphinx config value ``django_settings`` is added via
    :meth:`~sphinx.application.Sphinx.add_config_value` and
    :meth:`~sphinxcontrib_django2.docstrings.setup_django` is called on the
    :event:`config-inited` event.

    :param app: The Sphinx application object
    :type app: ~sphinx.application.Sphinx
    """
    from .patches import patch_django_for_autodoc

    # When running, make sure Django doesn't execute querysets
    # Fix module paths for intersphinx mappings
    patch_django_for_autodoc()

    # Register custom event which can be emitted after Django has been set up
    app.add_event("django-configured")

    # Set default to environment variable to enable backwards compatibility
    app.add_config_value(
        "django_settings", os.environ.get("DJANGO_SETTINGS_MODULE"), True
    )

    # Set default of django_show_db_tables to False
    app.add_config_value("django_show_db_tables", False, True)
    # Setup Django after config is initialized
    app.connect("config-inited", setup_django)

    # Load sphinx.ext.autodoc extension before registering events
    app.setup_extension("sphinx.ext.autodoc")

    # Generate docstrings for Django model fields
    # Register the docstring processor with sphinx
    app.connect("autodoc-process-docstring", improve_docstring)

    # influence skip rules
    app.connect("autodoc-skip-member", autodoc_skip)

    return {
        "version:": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


def setup_django(app, config):
    """
    This function calls :func:`django.setup` so it doesn't have to be done in the app's
    ``conf.py``.

    Called on the :event:`config-inited` event.

    :param app: The Sphinx application object
    :type app: ~sphinx.application.Sphinx

    :param config: The Sphinx configuration
    :type config: ~sphinx.config.Config

    :raises ~sphinx.errors.ConfigError: If setting ``django_settings`` is not set correctly
    """
    if not config.django_settings:
        raise ConfigError(
            "Please specify your Django settings in the configuration 'django_settings' in your "
            "conf.py"
        )
    try:
        importlib.import_module(config.django_settings)
    except ModuleNotFoundError as e:
        raise ConfigError(
            "The module you specified in the configuration 'django_settings' in your conf.py "
            "cannot be imported. Make sure the module path is correct and the source directoy is "
            "added to sys.path."
        ) from e
    os.environ["DJANGO_SETTINGS_MODULE"] = config.django_settings
    django.setup()

    # Emit event to allow code which depends on Django to run
    app.emit("django-configured")


def autodoc_skip(app, what, name, obj, skip, options):
    """
    Hook to tell autodoc to include or exclude certain fields (see :event:`autodoc-skip-member`).

    Sadly, it doesn't give a reference to the parent object,
    so only the ``name`` can be used for referencing.

    :param app: The Sphinx application object
    :type app: ~sphinx.application.Sphinx

    :param what: The parent type, ``class`` or ``module``
    :type what: str

    :param name: The name of the child method/attribute.
    :type name: str

    :param obj: The child value (e.g. a method, dict, or module reference)
    :type obj: object

    :param options: The current autodoc settings.
    :type options: dict
    """
    if name in EXCLUDE_MEMBERS:
        return True

    if name in INCLUDE_MEMBERS:
        return False

    return skip


def improve_docstring(app, what, name, obj, options, lines):
    """
    Hook to improve the autodoc docstrings for Django models
    (see :event:`autodoc-process-docstring`).

    :param what: The type of the object which the docstring belongs to (one of ``module``,
                 ``class``, ``exception``, ``function``, ``method`` and ``attribute``)
    :type what: str

    :param name: The fully qualified name of the object
    :type name: str

    :param obj: The documented object
    :type obj: object

    :param options: The options given to the directive: an object with attributes
                    ``inherited_members``, ``undoc_members``, ``show_inheritance`` and ``noindex``
                    that are ``True`` if the flag option of same name was given to the auto
                    directive
    :type options: object

    :param lines: A list of strings – the lines of the processed docstring – that the event
                  handler can modify in place to change what Sphinx puts into the output.
    :type lines: list [ str ]

    :return: The modified list of lines
    :rtype: list [ str ]
    """
    if what == "class":
        improve_class_docstring(app, obj, lines)
    elif what == "attribute":
        improve_attribute_docstring(obj, name, lines)
    elif what == "method":
        improve_method_docstring(name, lines)
    elif what == "data":
        improve_data_docstring(obj, lines)

    # Return the extended docstring
    return lines
